import os
import os.path
import stat
import time
import typing

from wheezy.template.engine import Engine
from wheezy.template.typing import Loader, SupportsRender


class FileLoader(Loader):
    """Loads templates from file system.

    ``directories`` - search path of directories to scan for template.
    ``encoding`` - decode template content per encoding.
    """

    def __init__(
        self, directories: typing.List[str], encoding: str = "UTF-8"
    ) -> None:
        searchpath: typing.List[str] = []
        for path in directories:
            abspath = os.path.abspath(path)
            assert os.path.exists(abspath)
            assert os.path.isdir(abspath)
            searchpath.append(abspath)
        self.searchpath = searchpath
        self.encoding = encoding

    def list_names(self) -> tuple[str, ...]:
        """Return a list of names relative to directories. Ignores any files
        and directories that start with dot.
        """
        names = []
        for path in self.searchpath:
            pathlen = len(path) + 1
            for dirpath, dirnames, filenames in os.walk(path):
                for i in [
                    i
                    for i, name in enumerate(dirnames)
                    if name.startswith(".")
                ]:
                    del dirnames[i]
                for filename in filenames:
                    if filename.startswith("."):
                        continue
                    name = os.path.join(dirpath, filename)[pathlen:]
                    name = name.replace("\\", "/")
                    names.append(name)
        return tuple(sorted(names))

    def get_fullname(self, name: str) -> typing.Optional[str]:
        """Returns a full path by a template name."""
        for path in self.searchpath:
            filename = os.path.join(path, name)
            if not os.path.exists(filename):
                continue
            if not os.path.isfile(filename):
                continue
            return filename
        else:
            return None

    def load(self, name: str) -> typing.Optional[str]:
        """Loads a template by name from file system."""
        filename = self.get_fullname(name)
        if filename:
            f = open(filename, "rb")
            try:
                return f.read().decode(self.encoding)
            finally:
                f.close()
        return None


class DictLoader(Loader):
    """Loads templates from python dictionary.

    ``templates`` - a dict where key corresponds to template name and
    value to template content.
    """

    def __init__(self, templates: typing.Mapping[str, str]) -> None:
        self.templates = templates

    def list_names(self) -> tuple[str, ...]:
        """List all keys from internal dict."""
        return tuple(sorted(self.templates.keys()))

    def load(self, name: str) -> typing.Optional[str]:
        """Returns template by name."""
        if name not in self.templates:
            return None
        return self.templates[name]


class ChainLoader(Loader):
    """Loads templates from ``loaders`` until first succeed."""

    def __init__(self, loaders: typing.List[Loader]) -> None:
        self.loaders = loaders

    def list_names(self) -> tuple[str, ...]:
        """Returns as list of names from all loaders."""
        names = set()
        for loader in self.loaders:
            names |= set(loader.list_names())
        return tuple(sorted(names))

    def load(self, name: str) -> typing.Optional[str]:
        """Returns template by name from the first loader that succeed."""
        for loader in self.loaders:
            source = loader.load(name)
            if source is not None:
                return source
        return None


class PreprocessLoader(Loader):
    """Performs preprocessing of loaded template."""

    def __init__(
        self,
        engine: Engine,
        ctx: typing.Optional[typing.Mapping[str, typing.Any]] = None,
    ) -> None:
        self.engine = engine
        self.ctx = ctx or {}

    def list_names(self) -> tuple[str, ...]:
        return self.engine.loader.list_names()

    def load(self, name: str) -> str:
        return self.engine.render(name, self.ctx, {}, {})


def autoreload(engine: Engine, enabled: bool = True) -> Engine:
    """Auto reload template if changes are detected in file.

    Limitation: master (inherited), imported and preprocessed templates.

    It is recommended to use application server that supports
    file reload instead.
    """
    if not enabled:
        return engine
    return AutoReloadProxy(engine)


# region: internal details


class AutoReloadProxy(Engine):
    def __init__(self, engine: Engine):
        from warnings import warn

        self.engine = engine
        self.names: typing.Dict[str, float] = {}
        warn(
            "autoreload limitation: master (inherited), imported "
            "and preprocessed templates. It is recommended to use "
            "application server that supports file reload instead.",
            stacklevel=3,
        )

    def get_template(self, name: str) -> SupportsRender:
        if self.file_changed(name):
            self.remove(name)
        return self.engine.get_template(name)

    def render(
        self,
        name: str,
        ctx: typing.Mapping[str, typing.Any],
        local_defs: typing.Mapping[str, typing.Any],
        super_defs: typing.Mapping[str, typing.Any],
    ) -> str:
        if self.file_changed(name):
            self.remove(name)
        return self.engine.render(name, ctx, local_defs, super_defs)

    def remove(self, name: str) -> None:
        self.engine.remove(name)

    # region: internal details

    def __getattr__(self, name: str) -> typing.Any:
        return getattr(self.engine, name)

    def file_changed(self, name: str) -> bool:
        try:
            last_known_stamp = self.names[name]
            current_time = time.time()
            if current_time - last_known_stamp <= 2:
                return False
        except KeyError:
            last_known_stamp = 0

        loader = self.engine.loader
        abspath = loader.get_fullname(name)  # type: ignore[attr-defined]
        if not abspath:
            return False

        last_modified_stamp = os.stat(abspath)[stat.ST_MTIME]
        if last_modified_stamp <= last_known_stamp:
            return False
        self.names[name] = last_modified_stamp
        return True



import sys
import typing
from types import ModuleType

from wheezy.template.builder import SourceBuilder, builder_scan
from wheezy.template.comp import allocate_lock  # type: ignore[attr-defined]
from wheezy.template.compiler import Compiler
from wheezy.template.lexer import Lexer, lexer_scan
from wheezy.template.parser import Parser, parser_scan
from wheezy.template.typing import (
    Loader,
    RenderTemplate,
    SupportsRender,
    TemplateClass,
    Token,
)


class Template(SupportsRender):
    """Simple template class."""

    __slots__ = ("name", "render_template")

    def __init__(self, name: str, render_template: RenderTemplate) -> None:
        self.name = name
        self.render_template = render_template

    def render(self, ctx: typing.Mapping[str, typing.Any]) -> str:
        return self.render_template(ctx, {}, {})


class Engine(object):
    """The core component of template engine."""

    def __init__(
        self,
        loader: Loader,
        extensions: typing.List[typing.Any],
        template_class: typing.Optional[TemplateClass] = None,
    ) -> None:
        self.lock = allocate_lock()
        self.templates: typing.Dict[str, SupportsRender] = {}
        self.renders: typing.Dict[str, RenderTemplate] = {}
        self.modules: typing.Dict[str, ModuleType] = {}
        self.global_vars = {"_r": self.render, "_i": self.import_name}
        self.loader = loader
        self.template_class = template_class or Template
        self.compiler = Compiler(
            self.global_vars, sys.version_info >= (3, 11, 0) and -1 or -2
        )
        self.lexer = Lexer(**lexer_scan(extensions))
        self.parser = Parser(**parser_scan(extensions))
        self.builder = SourceBuilder(**builder_scan(extensions))

    def get_template(self, name: str) -> SupportsRender:
        """Returns compiled template."""
        try:
            return self.templates[name]
        except KeyError:
            self.compile_template(name)
            return self.templates[name]

    def render(
        self,
        name: str,
        ctx: typing.Mapping[str, typing.Any],
        local_defs: typing.Mapping[str, typing.Any],
        super_defs: typing.Mapping[str, typing.Any],
    ) -> str:
        """Renders template by name in given context."""
        try:
            return self.renders[name](ctx, local_defs, super_defs)
        except KeyError:
            self.compile_template(name)
            return self.renders[name](ctx, local_defs, super_defs)

    def remove(self, name: str) -> None:
        """Removes given ``name`` from internal cache."""
        self.lock.acquire(True)
        try:
            if name in self.renders:
                del self.templates[name]
                del self.renders[name]
            if name in self.modules:
                del self.modules[name]
        finally:
            self.lock.release()

    # region: internal details

    def import_name(self, name: str) -> ModuleType:
        try:
            return self.modules[name]
        except KeyError:
            self.compile_import(name)
            return self.modules[name]

    def compile_template(self, name: str) -> None:
        self.lock.acquire(True)
        try:
            if name not in self.renders:
                template_source = self.loader.load(name)
                if template_source is None:
                    raise IOError('Template "%s" not found.' % name)
                tokens = self.lexer.tokenize(template_source)
                nodes = self.parser.parse(tokens)
                source = self.builder.build_render(nodes)

                # print_debug(name, tokens, nodes, source)

                try:
                    render_template = self.compiler.compile_source(
                        source, name
                    )["render"]
                except SyntaxError as e:
                    raise complement_syntax_error(e, template_source, source)

                self.renders[name] = render_template
                self.templates[name] = self.template_class(
                    name, render_template
                )
        finally:
            self.lock.release()

    def compile_import(self, name: str) -> None:
        self.lock.acquire(True)
        try:
            if name not in self.modules:
                template_source = self.loader.load(name)
                if template_source is None:
                    raise IOError('Import "%s" not found.' % name)
                tokens = self.lexer.tokenize(template_source)
                nodes = self.parser.parse(tokens)
                source = self.builder.build_module(nodes)

                # print_debug(name, tokens, nodes, source)

                try:
                    self.modules[name] = self.compiler.compile_module(
                        source, name
                    )
                except SyntaxError as e:
                    raise complement_syntax_error(e, template_source, source)
        finally:
            self.lock.release()


# region: internal details


def print_debug(
    name: str,
    tokens: typing.List[Token],
    nodes: typing.List[typing.Any],
    source: str,
) -> None:  # pragma: nocover
    print(name.center(80, "-"))
    from pprint import pprint

    # pprint(tokens)
    pprint(nodes)
    from wheezy.template.utils import print_source

    print_source(source, -1)


def complement_syntax_error(
    err: SyntaxError, template_source: str, source: str
) -> SyntaxError:
    """Complements SyntaxError with template and source snippets,
    like one below:

    .. code-block:: none

        File "shared/snippet/widget.html", line 4
            if :

        template snippet:
        02 <h1>
        03     @msg!h
        04     @if :
        05         sd
        06     @end

        generated snippet:
        02     _b = []; w = _b.append; w('<h1>\\n    ')
        03     w(h(msg)); w('\\n')
        04     if :
        05         w('        sd\\n')
        06

            if :
            ^
        SyntaxError: invalid syntax

    """
    lineno = err.lineno or 0
    text = """\
%s
template snippet:
%s

generated snippet:
%s

    %s""" % (
        err.text,
        source_chunk(template_source, lineno - 2, 1),
        source_chunk(source, lineno, -1),
        err.text and err.text.strip() or "",
    )
    return err.__class__(err.msg, (err.filename, lineno - 2, err.offset, text))


def source_chunk(source: str, lineno: int, offset: int, extra: int = 2) -> str:
    lines = source.split("\n", lineno + extra)
    s = max(0, lineno - extra - 1)
    e = min(len(lines), lineno + extra)
    r = []
    for i in range(s, e):
        line = lines[i]
        r.append("  %02d %s" % (i + offset, line))
    return "\n".join(r)



import ast
import typing
from _thread import allocate_lock  # noqa

Tuple: typing.Any = tuple
List: typing.Any = list


def adjust_source_lineno(source: str, name: str, lineno: int) -> typing.Any:
    node = compile(source, name, "exec", ast.PyCF_ONLY_AST)
    ast.increment_lineno(node, lineno)
    return node


__all__ = (
    "adjust_source_lineno",
    "Tuple",
    "List",
)



import typing
from abc import abstractmethod
from typing import TypeAlias

# flake8: noqa: E704

Token: TypeAlias = tuple[int, str, str]


class Builder:
    lineno: int

    @abstractmethod
    def start_block(self) -> None: ...  # pragma: nocover

    @abstractmethod
    def end_block(self) -> None: ...  # pragma: nocover

    @abstractmethod
    def add(self, lineno: int, code: str) -> None: ...  # pragma: nocover

    @abstractmethod
    def build_block(
        self, nodes: typing.Iterable[Token]
    ) -> None: ...  # pragma: nocover

    @abstractmethod
    def build_token(
        self,
        lineno: int,
        token: str,
        value: typing.Union[str, typing.Iterable[Token]],
    ) -> None: ...  # pragma: nocover


Tokenizer = typing.Callable[[typing.Match[str]], Token]
LexerRule = tuple[typing.Pattern[str], Tokenizer]
PreProcessorRule = typing.Callable[[str], str]
PostProcessorRule = typing.Callable[[list[Token]], str]
BuilderRule = typing.Callable[
    [
        Builder,
        int,
        str,
        typing.Union[str, list[str], typing.Iterable[Token], None],
    ],
    bool,
]
ParserRule = typing.Callable[[str], typing.Union[str, list[str]]]


class ParserConfig:
    end_tokens: list[str]
    continue_tokens: list[str]
    compound_tokens: list[str]
    out_tokens: list[str]


RenderTemplate = typing.Callable[
    [
        typing.Mapping[str, typing.Any],
        typing.Mapping[str, typing.Any],
        typing.Mapping[str, typing.Any],
    ],
    str,
]


class SupportsRender:
    @abstractmethod
    def render(
        self, ctx: typing.Mapping[str, typing.Any]
    ) -> str: ...  # pragma: nocover


TemplateClass = typing.Callable[[str, RenderTemplate], SupportsRender]


class Loader:
    @abstractmethod
    def list_names(self) -> tuple[str, ...]: ...  # pragma: nocover

    @abstractmethod
    def load(self, name: str) -> typing.Optional[str]: ...  # pragma: nocover



import typing

from wheezy.template.typing import (
    LexerRule,
    PostProcessorRule,
    PreProcessorRule,
    Token,
)


def lexer_scan(
    extensions: typing.List[typing.Any],
) -> typing.Mapping[str, typing.Any]:
    """Scans extensions for ``lexer_rules`` and ``preprocessors``
    attributes.
    """
    lexer_rules: typing.Dict[int, LexerRule] = {}
    preprocessors: typing.List[PreProcessorRule] = []
    postprocessors: typing.List[PostProcessorRule] = []
    for extension in extensions:
        if hasattr(extension, "lexer_rules"):
            lexer_rules.update(extension.lexer_rules)
        if hasattr(extension, "preprocessors"):
            preprocessors.extend(extension.preprocessors)
        if hasattr(extension, "postprocessors"):
            postprocessors.extend(extension.postprocessors)
    return {
        "lexer_rules": [lexer_rules[k] for k in sorted(lexer_rules.keys())],
        "preprocessors": preprocessors,
        "postprocessors": postprocessors,
    }


class Lexer(object):
    """Tokenizes input source per rules supplied."""

    def __init__(
        self,
        lexer_rules: typing.List[LexerRule],
        preprocessors: typing.Optional[typing.List[PreProcessorRule]] = None,
        postprocessors: typing.Optional[typing.List[PostProcessorRule]] = None,
        **ignore: typing.Any,
    ) -> None:
        """Initializes with ``rules``. Rules must be a list of
        two elements tuple: ``(regex, tokenizer)`` where
        tokenizer if a callable of the following contract::

        def tokenizer(match):
            return end_index, token, value
        """
        self.rules = lexer_rules
        self.preprocessors = preprocessors or []
        self.postprocessors = postprocessors or []

    def tokenize(self, source: str) -> typing.List[Token]:
        """Translates ``source`` accoring to lexer rules into
        an iteratable of tokens.
        """
        for preprocessor in self.preprocessors:
            source = preprocessor(source)
        tokens: typing.List[Token] = []
        append = tokens.append
        pos = 0
        lineno = 1
        end = len(source)
        while pos < end:
            for regex, tokenizer in self.rules:
                m = regex.match(source, pos, end)
                if m is not None:
                    npos, token, value = tokenizer(m)
                    assert npos > pos
                    append((lineno, token, value))
                    lineno += source[pos:npos].count("\n")
                    pos = npos
                    break
            else:
                raise AssertionError("Lexer pattern mismatch.")
        for postprocessor in self.postprocessors:
            postprocessor(tokens)
        return tokens



def find_all_balanced(text: str, start: int = 0) -> int:
    """Finds balanced ``([`` with ``])`` assuming
    that ``start`` is pointing to ``(`` or ``[`` in ``text``.
    """
    if start >= len(text) or text[start] not in "([":
        return start
    while 1:
        pos = find_balanced(text, start)
        pos = find_balanced(text, pos, "[", "]")
        if pos != start:
            start = pos
        else:
            return pos


def find_balanced(
    text: str, start: int = 0, start_sep: str = "(", end_sep: str = ")"
) -> int:
    """Finds balanced ``start_sep`` with ``end_sep`` assuming
    that ``start`` is pointing to ``start_sep`` in ``text``.
    """
    if start >= len(text) or start_sep != text[start]:
        return start
    balanced = 1
    pos = start + 1
    while pos < len(text):
        token = text[pos]
        pos += 1
        if token == end_sep:
            if balanced == 1:
                return pos
            balanced -= 1
        elif token == start_sep:
            balanced += 1
    return start


def print_source(source: str, lineno: int = 1) -> None:  # pragma: nocover
    lines = []
    for line in source.split("\n"):
        lines.append("%02d " % lineno + line)
        lineno += line.count("\n") + 1
    print("\n".join(lines))



import getopt
import json
import os
import sys
import typing

from wheezy.template.engine import Engine
from wheezy.template.ext.code import CodeExtension
from wheezy.template.ext.core import CoreExtension
from wheezy.template.loader import FileLoader

try:
    from wheezy.html.utils import escape_html as escape
except ImportError:  # pragma: nocover
    from html import escape


class Options:
    def __init__(
        self,
        token_start: str = "@",
        line_join: str = "\\",
        searchpath: typing.Optional[typing.List[str]] = None,
        extensions: typing.Optional[typing.List[str]] = None,
    ) -> None:
        self.token_start = token_start
        self.line_join = line_join
        self.searchpath = searchpath or ["."]
        self.extensions = extensions or []
        self.template = ""
        self.context: typing.List[str] = []


def main(argv: typing.Optional[typing.List[str]] = None) -> int:
    args = parse_args(argv or sys.argv[1:])
    if not args:
        return 2
    ts = args.token_start
    extensions = [CoreExtension(ts, args.line_join), CodeExtension(ts)]
    extensions.extend(args.extensions)
    engine = Engine(FileLoader(args.searchpath), extensions)
    engine.global_vars.update({"h": escape})
    t = engine.get_template(args.template)
    sys.stdout.write(t.render(load_context(args.context)))
    return 0


def load_context(sources: typing.List[str]) -> typing.Mapping[str, typing.Any]:
    c: typing.Dict[str, typing.Any] = {}
    args: typing.List[typing.Any] = []
    for s in sources:
        if os.path.isfile(s):
            d = json.load(open(s))
        else:
            d = json.loads(s)
        args.append(d)
        if isinstance(d, dict):
            c.update(d)
    c["__args__"] = args
    return c


def parse_args(  # noqa: C901
    args: typing.List[str],
) -> typing.Optional[Options]:
    try:
        opts, value = getopt.getopt(args, "s:t:j:wh")
    except getopt.GetoptError:
        e = sys.exc_info()[1]
        usage()
        print("error: %s" % e)
        return None
    d = Options(token_start="@", searchpath=["."], extensions=[])
    for o, a in opts:
        if o == "-h":
            usage()
            return None
        elif o == "-t":
            d.token_start = a
        elif o == "-j":
            d.line_join = a
        elif o == "-s":
            d.searchpath = a.split(";")
        elif o == "-w":  # pragma: nocover
            from wheezy.html.ext.template import WhitespaceExtension

            d.extensions.append(WhitespaceExtension())
    if not value:
        usage()
        return None
    d.template = value[0]
    d.context = value[1:]
    return d


def usage() -> None:
    from os.path import basename

    print(
        """\

Renders a template with the provided context.

Usage: %s template [ context ... ]

Positional arguments:
  template    The template filename.
  context     A filename or JSON string representing the context.

Optional arguments:
  -s path     Search path for templates (default ".").
  -t token    Token start (default "@").
  -j token    Line join token (default "\\").
  -w          Enable whitespace cleanup.
  -h          Show this help message.

The contexts passed are available as the __args__ list variable."""
        % basename(sys.argv[0])
    )


if __name__ == "__main__":  # pragma: nocover
    sys.exit(main())



import re
import typing

from wheezy.template.utils import find_balanced

RE_ARGS = re.compile(r'\s*(?P<expr>(([\'"]).*?\3|.+?))\s*\,')
RE_KWARGS = re.compile(
    r'\s*(?P<name>\w+)\s*=\s*(?P<expr>([\'"].*?[\'"]|.+?))\s*\,'
)
RE_STR_VALUE = re.compile(r'^[\'"](?P<value>.+)[\'"]$')
RE_INT_VALUE = re.compile(r"^(?P<value>(\d+))$")


# region: core extension


class DeterminedExtension(object):
    """Tranlates function calls between template engines.

    Strictly determined known calls are converted to preprocessor
    calls, e.g.::

        @_('Name:')
        @path_for('default')
        @path_for('static', path='/static/css/site.css')

    Those that are not strictly determined are ignored and processed
    by runtime engine.
    """

    def __init__(
        self,
        known_calls: typing.List[str],
        runtime_token_start: str = "@",
        token_start: str = "#",
    ) -> None:
        self.token_start = token_start
        self.pattern = re.compile(
            r"%s(%s)(?=\()" % (runtime_token_start, "|".join(known_calls))
        )
        self.preprocessors = [self.preprocess]

    def preprocess(self, source: str) -> str:
        result = []
        start = 0
        for m in self.pattern.finditer(source):
            pstart = m.end()
            pend = find_balanced(source, pstart)
            if determined(source[pstart + 1 : pend - 1]):
                name = m.group(1)
                result.append(source[start : m.start()])
                result.append(self.token_start + "ctx['" + name + "']")
                start = pstart
        if start:
            result.append(source[start:])
            return "".join(result)
        else:
            return source


def determined(expression: str) -> bool:
    """Checks if expresion is strictly determined.

    >>> determined("'default'")
    True
    >>> determined('name')
    False
    >>> determined("'default', id=id")
    False
    >>> determined("'default', lang=100")
    True
    >>> determined('')
    True
    """
    args, kwargs = parse_params(expression)
    for arg in args:
        if not str_or_int(arg):
            return False
    for arg in kwargs.values():
        if not str_or_int(arg):
            return False
    return True


def parse_kwargs(text: str) -> typing.Mapping[str, str]:
    """Parses key-value type of parameters.

    >>> parse_kwargs('id=item.id')
    {'id': 'item.id'}
    >>> sorted(parse_kwargs('lang="en", id=12').items())
    [('id', '12'), ('lang', '"en"')]
    """
    kwargs = {}
    for m in RE_KWARGS.finditer(text + ","):
        groups = m.groupdict()
        kwargs[groups["name"].rstrip("_")] = groups["expr"]
    return kwargs


def parse_args(text: str) -> typing.List[str]:
    """Parses argument type of parameters.

    >>> parse_args('')
    []
    >>> parse_args('10, "x"')
    ['10', '"x"']
    >>> parse_args("'x', 100")
    ["'x'", '100']
    >>> parse_args('"default"')
    ['"default"']
    """
    args = []
    for m in RE_ARGS.finditer(text + ","):
        args.append(m.group("expr"))
    return args


def parse_params(
    text: str,
) -> tuple[typing.List[str], typing.Mapping[str, str]]:
    """Parses function parameters.

    >>> parse_params('')
    ([], {})
    >>> parse_params('id=item.id')
    ([], {'id': 'item.id'})
    >>> parse_params('"default"')
    (['"default"'], {})
    >>> parse_params('"default", lang="en"')
    (['"default"'], {'lang': '"en"'})
    """
    if "=" in text:
        args = text.split("=")[0]
        if "," in args:
            args = args.rsplit(",", 1)[0]
            kwargs = text[len(args) :]
            return parse_args(args), parse_kwargs(kwargs)
        else:
            return [], parse_kwargs(text)
    else:
        return parse_args(text), {}


def str_or_int(text: str) -> bool:
    """Ensures ``text`` as string or int expression.

    >>> str_or_int('"default"')
    True
    >>> str_or_int("'Hello'")
    True
    >>> str_or_int('100')
    True
    >>> str_or_int('item.id')
    False
    """
    m = RE_STR_VALUE.match(text)
    if m:
        return True
    else:
        m = RE_INT_VALUE.match(text)
        if m:
            return True
    return False



import typing

from wheezy.template.typing import ParserConfig, ParserRule, Token


def parser_scan(
    extensions: typing.List[typing.Any],
) -> typing.Mapping[str, typing.Any]:
    parser_rules = {}
    parser_configs = []
    for extension in extensions:
        if hasattr(extension, "parser_rules"):
            parser_rules.update(extension.parser_rules)
        if hasattr(extension, "parser_configs"):
            parser_configs.extend(extension.parser_configs)
    return {
        "parser_rules": parser_rules,
        "parser_configs": parser_configs,
    }


class Parser(ParserConfig):
    """
    ``continue_tokens`` are used to insert ``end`` node right
    before them to simulate a block end. Such nodes have token
    value ``None``.

    ``out_tokens`` are combined together into a single node.
    """

    def __init__(
        self,
        parser_rules: typing.Dict[str, ParserRule],
        parser_configs: typing.Optional[
            typing.List[typing.Callable[[ParserConfig], None]]
        ] = None,
        **ignore: typing.Any,
    ) -> None:
        self.end_tokens: typing.List[str] = []
        self.continue_tokens: typing.List[str] = []
        self.compound_tokens: typing.List[str] = []
        self.out_tokens: typing.List[str] = []
        self.rules = parser_rules
        if parser_configs:
            for config in parser_configs:
                config(self)

    def end_continue(
        self, tokens: typing.List[Token]
    ) -> typing.Iterator[Token]:
        """If token is in ``continue_tokens`` prepend it
        with end token so it simulate a closed block.
        """
        for t in tokens:
            if t[1] in self.continue_tokens:
                yield (t[0], "end", "")
            yield t

    def parse_iter(
        self, tokens: typing.Iterator[Token]
    ) -> typing.Iterator[typing.Any]:
        operands = []
        for lineno, token, value in tokens:
            if token in self.rules:
                value = self.rules[token](value)  # type: ignore[assignment]
            if token in self.out_tokens:
                operands.append((lineno, token, value))
            else:
                if operands:
                    yield operands[0][0], "out", operands
                    operands = []
                if token in self.compound_tokens:
                    yield lineno, token, (value, list(self.parse_iter(tokens)))
                else:
                    yield lineno, token, value
                    if token in self.end_tokens:
                        break
        if operands:
            yield operands[0][0], "out", operands

    def parse(self, tokens: typing.List[Token]) -> typing.List[typing.Any]:
        return list(self.parse_iter(self.end_continue(tokens)))



import typing

from wheezy.template.comp import allocate_lock  # type: ignore[attr-defined]
from wheezy.template.engine import Engine
from wheezy.template.loader import ChainLoader, DictLoader
from wheezy.template.typing import Loader, SupportsRender


class Preprocessor(object):
    """Preprocess templates with ``engine`` and vary runtime templates
    by ``key_factory`` function using ``runtime_engine_factory``.
    """

    def __init__(
        self,
        runtime_engine_factory: typing.Callable[[Loader], Engine],
        engine: Engine,
        key_factory: typing.Callable[[typing.Mapping[str, typing.Any]], str],
    ) -> None:
        self.lock = allocate_lock()
        self.runtime_engines: typing.Dict[str, Engine] = {}
        self.runtime_engine_factory = runtime_engine_factory
        self.engine = engine
        self.loader = engine.loader
        self.key_factory = key_factory
        template_class = self.engine.template_class
        self.engine.template_class = lambda name, _: template_class(
            name,
            lambda ctx, local_defs, super_defs: self.render(
                name, ctx, local_defs, super_defs
            ),
        )

    def get_template(self, name: str) -> SupportsRender:
        return self.engine.get_template(name)

    def render(
        self,
        name: str,
        ctx: typing.Mapping[str, typing.Any],
        local_defs: typing.Mapping[str, typing.Any],
        super_defs: typing.Mapping[str, typing.Any],
    ) -> str:
        try:
            runtime_engine = self.runtime_engines[self.key_factory(ctx)]
        except KeyError:
            runtime_engine = self.ensure_runtime_engine(self.key_factory(ctx))
        try:
            return runtime_engine.renders[name](ctx, local_defs, super_defs)
        except KeyError:
            self.preprocess_template(runtime_engine, name, ctx)
            return runtime_engine.renders[name](ctx, local_defs, super_defs)

    def remove(self, name: str) -> None:
        self.lock.acquire(True)
        try:
            self.engine.remove(name)
            for runtime_engine in self.runtime_engines.values():
                runtime_engine.remove(name)
        finally:
            self.lock.release()

    # region: internal details

    def ensure_runtime_engine(self, key: str) -> Engine:
        self.lock.acquire(True)
        try:
            engines = self.runtime_engines
            if key in engines:  # pragma: nocover
                return engines[key]
            engine = engines[key] = self.runtime_engine_factory(
                ChainLoader([DictLoader({}), self.engine.loader])
            )

            def render(
                name: str,
                ctx: typing.Mapping[str, typing.Any],
                local_defs: typing.Mapping[str, typing.Any],
                super_defs: typing.Mapping[str, typing.Any],
            ) -> str:
                try:
                    return engine.renders[name](ctx, local_defs, super_defs)
                except KeyError:
                    self.preprocess_template(engine, name, ctx)
                    return engine.renders[name](ctx, local_defs, super_defs)

            engine.global_vars["_r"] = render
            return engine
        finally:
            self.lock.release()

    def preprocess_template(
        self,
        runtime_engine: Engine,
        name: str,
        ctx: typing.Mapping[str, typing.Any],
    ) -> None:
        self.lock.acquire(True)
        try:
            if name not in runtime_engine.renders:
                source = self.engine.render(name, ctx, {}, {})
                loader = runtime_engine.loader.loaders[0]  # type: ignore
                loader.templates[name] = source
                runtime_engine.compile_template(name)
                del loader.templates[name]
        finally:
            self.lock.release()
